@@ -1,13 +1,20 @@
 import sys
+import os
 import types
 import ctypes
 import struct
 import numbers
 
+_DEBUGLEVEL = os.getenv ('LUA_DEBUG')
+if _DEBUGLEVEL:
+	_DEBUGLEVEL = int (_DEBUGLEVEL)
+else:
+	_DEBUGLEVEL = None
+
 def _dprint (text, level = 0):
 	'''Print stuff for debugging. This function does nothing when not debugging'''
 	# Set the following to the desired debugging level.
-	if level > 5:
+	if _DEBUGLEVEL is not None and level > _DEBUGLEVEL:
 		sys.stderr.write (text + '\n')
 
 # option for multiple returns in `lua_pcall' and `lua_call'
@@ -99,6 +106,7 @@ class lua:
 		self._the_put = self._factory (_object_put)
 		self._the_call = self._factory (_object_call)
 		self._the_gc = self._factory (_object_gc)
+		self._the_tostring = self._factory (_object_tostring)
 		lua._lib.lua_createtable (self._state, 0, 16)
 		lua._lib.lua_pushcclosure (self._state, self._the_add, 0)
 		lua._lib.lua_setfield (self._state, -2, '__add')
@@ -132,6 +140,8 @@ class lua:
 		lua._lib.lua_setfield (self._state, -2, '__call')
 		lua._lib.lua_pushcclosure (self._state, self._the_gc, 0)
 		lua._lib.lua_setfield (self._state, -2, '__gc')
+		lua._lib.lua_pushcclosure (self._state, self._the_tostring, 0)
+		lua._lib.lua_setfield (self._state, -2, '__tostring')
 		lua._lib.lua_setfield (self._state, LUA_REGISTRYINDEX, 'metatable')
 		if not debug:
 			self.run ('debug = nil', name = 'disabling debug')
@@ -186,13 +196,7 @@ class lua:
 					k = key
 				module[k] = getattr (value, key)
 			value = module
-		# This must be a real lua table, so don't use _push.
-		lua._lib.lua_createtable (self._state, 0, 0)
-		pos = lua._lib.lua_gettop (self._state)
-		for i in value:
-			self._push (i)
-			self._push (value[i])
-			lua._lib.lua_settable (self._state, pos)
+		self._push_luatable (value)
 		lua._lib.lua_setfield (self._state, LUA_GLOBALSINDEX, name)
 		self.run ('module ("%s")' % name, name = 'loading module %s' % name)
 	def _to_python (self, index):
@@ -223,7 +227,7 @@ class lua:
 		elif lua._lib.lua_iscfunction (self._state, index):
 			return lua._lib.lua_tocfunction (self._state, index)
 		else:
-			raise AssertionError ('unexpected lua type')
+			raise AssertionError ('unexpected lua type %d' % type)
 	def _push (self, obj):
 		_dprint ('pushing %s (%s) to lua stack' % (str (obj), type (obj)), 1)
 		if obj is None:
@@ -258,6 +262,23 @@ class lua:
 			# This shouldn't be possible: everything is an object.
 			raise ValueError ('object of type %s cannot be converted to lua object' % str (type (obj)))
 		_dprint ('done pushing', 1)
+	def _push_luatable (self, table):
+		'''Push a real lua table to the stack; not a userdata emulating a table. Table must be a list or dict'''
+		_dprint ('pushing lua table', 1)
+		if isinstance (table, dict):
+			lua._lib.lua_createtable (self._state, 0, len (table))
+			for i in table:
+				self._push (i)
+				self._push (table[i])
+				lua._lib.lua_settable (self._state, -3)
+		else:
+			lua._lib.lua_createtable (self._state, len (table), 0)
+			for i in range (len (table)):
+				self._push (table[i])
+				lua._lib.lua_rawseti (self._state, -2, i + 1)
+		_dprint ('done pushing lua table', 1)
+	def make_table (self, data):
+		return Table (self, data)
 
 def _object_add (state):
 	_dprint ('adding lua stuff', 1)
@@ -517,6 +538,13 @@ def _object_gc (state):
 	del self._objects[id]
 	return 0
 
+def _object_tostring (state):
+	_dprint ('tostringing lua stuff', 1)
+	self = lua._objects[state]
+	id = lua._lib.lua_touserdata (state, 1)
+	self._push (str (self._objects[id]))
+	return 1
+
 class Function:
 	def __init__ (self, luaobject):
 		_dprint ('creating lua function', 1)
@@ -537,19 +565,31 @@ class Function:
 	def __del__ (self):
 		_dprint ('destroying lua function', 1)
 		lua._lib.luaL_unref (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		#super (Function, self).__del__ (self)
 
 class Table:
-	def __init__ (self, luaobject):
+	def __init__ (self, luaobject, table = None):
 		_dprint ('creating lua table', 1)
 		self._luaobject = luaobject
+		if table is not None:
+			self._luaobject._push_luatable (table)
 		self._id = lua._lib.luaL_ref (self._luaobject._state, LUA_REGISTRYINDEX)
 	def __len__ (self):
 		_dprint ('requesting length of lua table', 3)
 		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
-		lua._lib.lua_objlen (self._luaobject._state, -1)
-		ret = self._luaobject._to_python (-1)
-		lua._lib.lua_settop (self._luaobject._state, -3)
+		ret = lua._lib.lua_objlen (self._luaobject._state, -1)
+		lua._lib.lua_settop (self._luaobject._state, -2)
 		return ret
+	def __iadd__ (self, other):
+		_dprint ('adding objects to lua table', 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		l = lua._lib.lua_objlen (self._luaobject._state, -1)
+		for i in other:
+			self._luaobject._push (i)
+			lua._lib.lua_rawseti (self._luaobject._state, -2, l + 1)
+			l += 1
+		lua._lib.lua_settop (self._luaobject._state, -2)
+		return self
 	def __getitem__ (self, key):
 		_dprint ('requesting item of lua table: %s' % key, 3)
 		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
@@ -557,6 +597,8 @@ class Table:
 		lua._lib.lua_gettable (self._luaobject._state, -2)
 		ret = self._luaobject._to_python (-1)
 		lua._lib.lua_settop (self._luaobject._state, -3)
+		if ret is None:
+			raise IndexError
 		return ret
 	def __setitem__ (self, key, value):
 		_dprint ('setting item of lua table: %s: %s' % (key, value), 3)
@@ -565,6 +607,11 @@ class Table:
 		self._luaobject._push (value)
 		lua._lib.lua_settable (self._luaobject._state, -3)
 		lua._lib.lua_settop (self._luaobject._state, -2)
+	def __delitem__ (self, key):
+		_dprint ('deleting item of lua table: %s' % key, 3)
+		# Raise IndexError if key doesn't exist: use getitem.
+		self[key]
+		self[key] = None
 	def __contains__ (self, key):
 		_dprint ('checking if %s is in lua table' % key, 3)
 		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
@@ -595,6 +642,7 @@ class Table:
 	def __del__ (self):
 		_dprint ('destroying lua table', 1)
 		lua._lib.luaL_unref (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		#super (Table, self).__del__ (self)
 	def __eq__ (self, other):
 		if not isinstance (other, Table):
 			return False
@@ -605,7 +653,7 @@ class Table:
 		return ret
 	def __ne__ (self, other):
 		return not self == other
-	def _dict (self):
+	def dict (self):
 		'Get a copy of the table as a dict'
 		_dprint ('get dict from lua table', 1)
 		ret = {}
@@ -616,7 +664,7 @@ class Table:
 			lua._lib.lua_settop (self._luaobject._state, -2)
 		lua._lib.lua_settop (self._luaobject._state, -2)
 		return ret
-	def _list (self):
+	def list (self):
 		'Get a copy of the table, which must be a sequence, as a list'
 		_dprint ('get list from lua table', 1)
 		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
@@ -628,3 +676,16 @@ class Table:
 			lua._lib.lua_settop (self._luaobject._state, -2)
 		lua._lib.lua_settop (self._luaobject._state, -2)
 		return ret
+	def pop (self, index = -1):
+		'''Like list.pop'''
+		_dprint ('popping item %d of lua table.' % index, 3)
+		lua._lib.lua_rawgeti (self._luaobject._state, LUA_REGISTRYINDEX, self._id)
+		if index < 0:
+			index += len (self)
+		lua._lib.lua_rawgeti (self._luaobject._state, -2, index + 1)
+		ret = self._luaobject._to_python (-1)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+		self._luaobject._push (None)
+		lua._lib.lua_rawseti (self._luaobject._state, -2, index + 1)
+		lua._lib.lua_settop (self._luaobject._state, -2)
+		return ret
